/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_cursor.c,v 1.6 1998/09/14 06:55:58 slouken Exp $";
#endif

/* General cursor handling code for SDL */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_mutex.h"
#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_mouse.h"
#include "SDL_blit.h"
#include "SDL_events_c.h"
#include "SDL_sysvideo.h"
#include "SDL_sysevents.h"
#include "SDL_cursor_c.h"
#include "SDL_pixels_c.h"
#include "SDL_video_c.h"
#include "SDL_sysmouse_c.h"
#include "default_cursor.h"

/* These are static for our cursor handling code */
int SDL_cursorstate = CURSOR_VISIBLE;
static int SDL_movingcursor = 0;
static SDL_Cursor *SDL_cursor = NULL;
static SDL_Rect mousesrc;
static SDL_Rect mousedst;
static SDL_Cursor *SDL_defcursor = NULL;

/* Lock the display against threads and draw cursor */
static void Lock_DrawCursor(SDL_Surface *screen);
static void Lock_EraseCursor(SDL_Surface *screen);


/* Public functions */
void SDL_CursorQuit(void)
{
	if ( SDL_cursor != NULL ) {
		SDL_Cursor *cursor;

		SDL_cursorstate &= ~CURSOR_VISIBLE;
		SDL_FreeCursor(SDL_cursor);
		if ( SDL_defcursor != NULL ) {
			cursor = SDL_defcursor;
			SDL_defcursor = NULL;
			SDL_FreeCursor(cursor);
		}
		SDL_cursor = NULL;
	}
}
int SDL_CursorInit(void)
{
	/* We don't have mouse focus, and the cursor isn't drawn yet */
	SDL_cursorstate = CURSOR_VISIBLE;
	SDL_movingcursor = 0;

	/* Create the default cursor */
	if ( SDL_defcursor == NULL ) {
		SDL_defcursor = SDL_CreateCursor(default_cdata, default_cmask,
					DEFAULT_CWIDTH, DEFAULT_CHEIGHT,
						DEFAULT_CHOTX, DEFAULT_CHOTY);
		SDL_SetCursor(SDL_defcursor);
	}

	/* That's it! */
	return(0);
}

/* Software cursor drawing support */
SDL_Cursor * SDL_CreateCursor (Uint8 *data, Uint8 *mask, 
					int w, int h, int hot_x, int hot_y)
{
	SDL_Cursor  *cursor;
	WMcursor    *wm_cursor;
	SDL_Surface *sw_cursor;
	SDL_Palette *palette;
	Uint8 *pixels;
	int row, col;
	Uint8  cbits=0, mbits=0;

	if ( ! SDL_CursorSurface->w || ! SDL_CursorSurface->h ) {
		SDL_SetError("You must set video mode before creating cursors");
		return(NULL);
	}

	/* Sanity check the hot spot */
	if ( (hot_x >= w) || (hot_y >= h) ) {
		SDL_SetError("Cursor hot spot doesn't lie within cursor");
		return(NULL);
	}

	/* If the window manager gives us a good cursor, we're done! */
	wm_cursor = SDL_CreateWMCursor(SDL_CursorSurface, data, mask,
							w, h, hot_x, hot_y);
	if ( wm_cursor != NULL ) {
		cursor = (SDL_Cursor *)malloc(sizeof(SDL_Cursor));
		if ( cursor == NULL ) {
			SDL_OutOfMemory();
			return(NULL);
		}
		cursor->wm_cursor = wm_cursor;
		cursor->sw_cursor = NULL;
		cursor->sw_erased = NULL;
		cursor->hot_x = hot_x;
		cursor->hot_y = hot_y;
		cursor->remap = 0;
		return(cursor);
	}
	/* Otherwise, either there was an unknown window manager error
	   or there is no window manager to provide a mouse cursor.
	 */

	/* Allocate a software cursor surface */
	sw_cursor = SDL_AllocSurface
				(SDL_SWSURFACE, w, h, 8, 0, 0, 0, 0);
	if ( sw_cursor == NULL ) {
		return(NULL);
	}

	/* Fill the surface with the cursor image */
	pixels = (Uint8 *)sw_cursor->pixels;
	palette = sw_cursor->format->palette;
	palette->colors[0].r = 0x00;
	palette->colors[0].g = 0x00;
	palette->colors[0].b = 0x00;
	palette->colors[1].r = 0xFF;
	palette->colors[1].g = 0xFF;
	palette->colors[1].b = 0xFF;
	palette->colors[2].r = 0x80;
	palette->colors[2].g = 0x80;
	palette->colors[2].b = 0x80;
	for ( row=0; row<h; ++row ) {
		for ( col=0; col<w; ++col ) {
			if ( (col%8) == 0 ) {
				cbits = *data++;
				mbits = *mask++;
			}
			if ( (mbits&0x80) == 0x80 ) {
				*pixels = !(cbits&0x80);	/* 0 or 1 */
			} else {
				*pixels = 2;
			}
			cbits <<= 1;
			mbits <<= 1;
			++pixels;
		}
	}
	/* Set the transparent pixel */
	SDL_SetColorKey(sw_cursor, SDL_SRCCOLORKEY, 2);

	/* Now it's just a cursor surface -- do it! */
	cursor = SDL_CreateColorCursor(sw_cursor, hot_x, hot_y);
	SDL_FreeSurface(sw_cursor);
	return(cursor);
}

SDL_Cursor * SDL_CreateColorCursor 
				(SDL_Surface *sw_cursor, int hot_x, int hot_y)
{
	SDL_Cursor  *cursor;

#ifdef THREADED_EVENTS
	/* Check to make sure cursor locking will work */
	if ( (SDL_CursorSurface->flags & SDL_THREADSAFE) != SDL_THREADSAFE ) {
		SDL_SetError("Display surface is not thread-safe!");
		return(NULL);
	}
#endif
	/* Allocate the software cursor */
	cursor = (SDL_Cursor *)malloc(sizeof(SDL_Cursor));
	if ( cursor == NULL ) {
		SDL_OutOfMemory();
		return(NULL);
	}

	/* No window manager cursor - always software when in living color */
	cursor->wm_cursor = NULL;
	cursor->hot_x = hot_x;
	cursor->hot_y = hot_y;
	cursor->remap = 0;

	/* Allocate the background of the cursor */
	cursor->sw_erased = SDL_AllocSurface
				((SDL_CursorSurface->flags&SDL_HWSURFACE),
				sw_cursor->w, sw_cursor->h,
				SDL_CursorSurface->format->BitsPerPixel,
					SDL_CursorSurface->format->Rmask,
					SDL_CursorSurface->format->Gmask,
					SDL_CursorSurface->format->Bmask, 0);
	if ( cursor->sw_erased == NULL ) {
		SDL_FreeCursor(cursor);
		return(NULL);
	}
	/* Calculate an identity mapping */
	cursor->sw_erased->identity = 1;
	cursor->sw_erased->dstfmt = SDL_CursorSurface->format;
	SDL_CalculateBlit(cursor->sw_erased);

	/* Convert the cursor surface to the video format */
	cursor->sw_cursor = SDL_ConvertSurface(sw_cursor,
				SDL_CursorSurface->format,
				(SDL_CursorSurface->flags&SDL_HWSURFACE));
	if ( cursor->sw_cursor == NULL ) {
		SDL_FreeCursor(cursor);
		return(NULL);
	}

	/* Use software blit acceleration if possible */
	if ( ((cursor->sw_cursor->flags & SDL_HWSURFACE) == SDL_SWSURFACE) &&
	     ((cursor->sw_cursor->flags & SDL_SRCCOLORKEY)==SDL_SRCCOLORKEY) &&
	     ((cursor->sw_cursor->flags & SDL_RLEACCEL) != SDL_RLEACCEL) ) {
#ifdef CURSOR_DEBUG
		fprintf(stderr,"Setting accelerated software blit on cursor\n");
#endif
		SDL_SetColorKey(cursor->sw_cursor, SDL_SRCCOLORKEY|SDL_RLEACCEL,
					cursor->sw_cursor->format->colorkey);
	}

	/* We're done! */
	return(cursor);
}

void SDL_SetCursor (SDL_Cursor *cursor)
{
	/* Set the new cursor */
	if ( cursor ) {
		/* Erase the current mouse position */
		if ( SHOULD_DRAWCURSOR(SDL_cursorstate) ) {
			Lock_EraseCursor(SDL_CursorSurface);
		}
		SDL_cursor = cursor;
	}

	/* Draw the new mouse cursor */
	if ( SDL_cursor && (SDL_cursorstate&CURSOR_VISIBLE) ) {
		/* Use window manager cursor if possible */
		if ( SDL_cursor->wm_cursor ) {
			SDL_cursorstate &= ~CURSOR_USINGSW;
			SDL_ShowWMCursor(SDL_CursorSurface,
						SDL_cursor->wm_cursor);
		} else {
			SDL_cursorstate |= CURSOR_USINGSW;
			SDL_ShowWMCursor(SDL_CursorSurface, NULL);
			mousesrc.w = mousedst.w = SDL_cursor->sw_cursor->w;
			mousesrc.h = mousedst.h = SDL_cursor->sw_cursor->h;
			if ( SHOULD_DRAWCURSOR(SDL_cursorstate) ) {
				Lock_DrawCursor(SDL_CursorSurface);
			}
		}
	} else {
		/* Erase window manager mouse (cursor not visible) */
		SDL_ShowWMCursor(SDL_CursorSurface, NULL);
	}
}

SDL_Cursor * SDL_GetCursor (void)
{
	return(SDL_cursor);
}

void SDL_FreeCursor (SDL_Cursor *cursor)
{
	if ( cursor == SDL_cursor ) {
		SDL_SetCursor(SDL_defcursor);
	}
	if ( cursor != SDL_defcursor ) {
		if ( cursor->sw_cursor ) {
			SDL_FreeSurface(cursor->sw_cursor);
			SDL_FreeSurface(cursor->sw_erased);
		}
		if ( cursor->wm_cursor ) {
			SDL_FreeWMCursor(SDL_CursorSurface, cursor->wm_cursor);
		}
		free(cursor);
	}
}

int SDL_ShowCursor (int toggle)
{
	int showing;

	SDL_cursorstate |= CURSOR_FOCUSED;
	showing = (SDL_cursorstate & CURSOR_VISIBLE);
	if ( toggle ) {
		SDL_cursorstate |= CURSOR_VISIBLE;
	} else {
		SDL_cursorstate &= ~CURSOR_VISIBLE;
	}
	if ( (SDL_cursorstate & CURSOR_VISIBLE) != showing ) {
		if ( (SDL_cursorstate & CURSOR_VISIBLE) ) {
			if ( (SDL_cursorstate & CURSOR_USINGSW) ) {
				Lock_DrawCursor(SDL_CursorSurface);
			} else {
				SDL_ShowWMCursor(SDL_CursorSurface,
							SDL_cursor->wm_cursor);
			}
		} else {
			if ( (SDL_cursorstate & CURSOR_USINGSW) ) {
				Lock_EraseCursor(SDL_CursorSurface);
			} else {
				SDL_ShowWMCursor(SDL_CursorSurface, NULL);
			}
		}
	}
	return(toggle);
}

void SDL_WarpMouse (Uint16 x, Uint16 y)
{
	/* This generates a mouse motion event */
	SDL_WarpWMCursor(SDL_CursorSurface, x, y);
}

void SDL_MouseFocus(int focus)
{
	if ( focus ) {
		SDL_cursorstate |= CURSOR_FOCUSED;
	} else {
		SDL_cursorstate &= ~CURSOR_FOCUSED;
	}
	if ( (SDL_cursorstate & CURSOR_USINGSW) && 
				(SDL_cursorstate & CURSOR_VISIBLE) ) {
		if ( focus ) {
			Lock_DrawCursor(SDL_CursorSurface);
		} else {
			Lock_EraseCursor(SDL_CursorSurface);
		}
	}
}

/* These rectangles are filled and clipped by SDL_DrawCursor():
	updates[0] = destination rectangle
	updates[1] = reserved
	updates[2] = source rectangle
 */
static SDL_Rect updates[3];
static int numupdates = 1;

void SDL_MoveCursor(Uint16 x, Uint16 y)
{
	/* Erase and update the current mouse position */
	SDL_cursorstate |= CURSOR_FOCUSED;
	if ( SHOULD_DRAWCURSOR(SDL_cursorstate) ) {
		if ( SDL_cursor ) {
#ifdef THREADED_EVENTS
			if ( SDL_mutexP(SDL_CursorSurface->mutex) == 0 ) {
				SDL_movingcursor = 1;
				SDL_EraseCursor(SDL_CursorSurface);
				updates[1] = updates[0];
				numupdates = 2;
				mousedst.x = x-SDL_cursor->hot_x;
				mousedst.y = y-SDL_cursor->hot_y;
				SDL_DrawCursor(SDL_CursorSurface);
				numupdates = 1;
				SDL_movingcursor = 0;
				SDL_mutexV(SDL_CursorSurface->mutex);
			} else {
				mousedst.x = x-SDL_cursor->hot_x;
				mousedst.y = y-SDL_cursor->hot_y;
			}
#else
			SDL_movingcursor = 1;
			SDL_EraseCursor(SDL_CursorSurface);
			updates[1] = updates[0];
			numupdates = 2;
			mousedst.x = x-SDL_cursor->hot_x;
			mousedst.y = y-SDL_cursor->hot_y;
			SDL_DrawCursor(SDL_CursorSurface);
			numupdates = 1;
			SDL_movingcursor = 0;
#endif /* THREADED_EVENTS */
		}
	} else {
		if ( SDL_cursor ) {
			mousedst.x = x-SDL_cursor->hot_x;
			mousedst.y = y-SDL_cursor->hot_y;
		}
	}
}
int SDL_OccludeCursor(SDL_Surface *dst, SDL_Rect *rect)
{
	if ( SDL_cursor && (dst == SDL_CursorSurface) && 
				SHOULD_DRAWCURSOR(SDL_cursorstate) ) {
		/* Return true if the rectangles overlap */
		if ( (SDL_cursor == NULL) ||
		     ((rect->x+rect->w) <= updates[0].x) ||
		     ((rect->y+rect->h) <= updates[0].y) ||
		     ((updates[0].x+updates[0].w) <= rect->x) ||
		     ((updates[0].y+updates[0].h) <= rect->y) ) {
			return(0);
		}
		return(1);
	} else {
		return(0);
	}
}
/* The colormap changed, so remap the cursor */
void SDL_RecolorCursor(void)
{
	if ( SDL_cursor == NULL ) {
		return;
	}
	SDL_cursor->remap = 1;	/* We're lazy :) */
}

static void Lock_EraseCursor(SDL_Surface *screen)
{
#ifdef THREADED_EVENTS
	if ( SDL_mutexP(screen->mutex) == 0 ) {
		SDL_EraseCursor(screen);
		SDL_mutexV(screen->mutex);
	}
#else
	SDL_EraseCursor(screen);
#endif
}
static void Lock_DrawCursor(SDL_Surface *screen)
{
#ifdef THREADED_EVENTS
	if ( SDL_mutexP(screen->mutex) == 0 ) {
		SDL_DrawCursor(screen);
		SDL_mutexV(screen->mutex);
	}
#else
	SDL_DrawCursor(screen);
#endif
}

void SDL_DrawCursor(SDL_Surface *screen)
{
	SDL_blit blit_cursor;
	int diff;

	/* Sanity check */
	if ( SDL_cursor == NULL )
		return;

	/* Check to see if the palette or screen format has changed */
	if ( SDL_cursor->remap ) {
		SDL_Surface *sw_cursor;

#ifdef CURSOR_DEBUG
		fprintf(stderr, "Remapping surface %p\n", 
						SDL_cursor->sw_cursor);
#endif
		sw_cursor = SDL_ConvertSurface(SDL_cursor->sw_cursor,
					screen->format,
					(screen->flags&SDL_HWSURFACE));
		if ( sw_cursor == NULL ) {
			return;
		}
		SDL_FreeSurface(SDL_cursor->sw_cursor);
		SDL_cursor->sw_cursor = sw_cursor;
		SDL_cursor->remap = 0;
	}

	/* Save important rectangles :-) */
	updates[2] = mousesrc;
	updates[0] = mousedst;

	/* Since we know that we're just copying pixels, we avoid having to
	   go through the blitting chain by performing our own clipping and
	   copying.  Note that we'll deadlock and infinitely recurse if we
	   call SDL_BlitSurface() here.
	 */

	if ( mousedst.x >= 0 ) {
		diff = (mousedst.x+mousedst.w)-screen->w;
		if ( diff > 0 ) {
			updates[2].w -= diff;
			updates[0].w -= diff;
		}
	} else {
		updates[2].x = -mousedst.x;
		updates[2].w += mousedst.x;
		updates[0].x = 0;
		updates[0].w += mousedst.x;
	}
	if ( mousedst.y >= 0 ) {
		diff = (mousedst.y+mousedst.h)-screen->h;
		if ( diff > 0 ) {
			updates[2].h -= diff;
			updates[0].h -= diff;
		}
	} else {
		updates[2].y = -mousedst.y;
		updates[2].h += mousedst.y;
		updates[0].y = 0;
		updates[0].h += mousedst.y;
	}
#ifdef CURSOR_DEBUG
	fprintf(stderr, 
		"Saving %dx%d from video (%d,%d) --> cursor (%d,%d)\n",
						updates[2].w, updates[2].h,
						updates[0].x, updates[0].y,
						updates[2].x, updates[2].y);
#endif
	/* Save the cursor area if it's not already saved */
	if ( 1 ) {
		SDL_Surface *sw_erased;
		Uint16 w, h;
		Uint8 *srcbuf, *dstbuf;

		/* Lock the surfaces and copy */
		sw_erased = SDL_cursor->sw_erased;
		if ( screen->flags & SDL_HWSURFACE ) {
			if ( SDL_HWCaps.LockHWSurface(screen) < 0 ) {
				return;
			}
		}
		if ( sw_erased->flags & SDL_HWSURFACE ) {
			if ( SDL_HWCaps.LockHWSurface(sw_erased) < 0 ) {
				if ( (screen->flags & SDL_HWSURFACE) ) {
					SDL_HWCaps.UnlockHWSurface(screen);
				}
				return;
			}
		}
		h = updates[0].h;
		w = updates[0].w*screen->format->BytesPerPixel;
		srcbuf = ((Uint8 *)screen->pixels + screen->offset +
				updates[0].y*screen->pitch +
				updates[0].x*screen->format->BytesPerPixel);
		dstbuf = ((Uint8 *)sw_erased->pixels + 
				updates[2].y*sw_erased->pitch +
				updates[2].x*sw_erased->format->BytesPerPixel);
		while ( h-- ) {
			/* We should really check to see if the pixels we
			   are reading are those of the mouse cursor and
			   ignore them if they are.  This happens when we
			   draw the cursor on Unlock(), but the area under
			   the cursor hasn't changed.  Slow, but fix?
			 */
			memcpy(dstbuf, srcbuf, w);
			srcbuf += screen->pitch;
			dstbuf += sw_erased->pitch;
		}
		if ( sw_erased->flags & SDL_HWSURFACE ) {
			SDL_HWCaps.UnlockHWSurface(sw_erased);
		}
		if ( screen->flags & SDL_HWSURFACE ) {
			SDL_HWCaps.UnlockHWSurface(screen);
		}
	}

	/* Blit the mouse cursor, update last and current position */
	if ( (SDL_cursor->sw_cursor->flags & SDL_HWACCEL) == SDL_HWACCEL ) {
		blit_cursor = SDL_cursor->sw_cursor->hw_blit;
	} else {
		blit_cursor = SDL_cursor->sw_cursor->sw_blit;
	}
	blit_cursor(SDL_cursor->sw_cursor, &updates[2], screen, &updates[0]);
	SDL_UpdateRects(screen, numupdates, updates);
	SDL_cursorstate |= CURSOR_ONSCREEN;
}

void SDL_EraseCursor(SDL_Surface *screen)
{
	SDL_blit blit_cursor;

	/* Sanity check */
	if ( (SDL_cursor == NULL) || ! (SDL_cursorstate & CURSOR_ONSCREEN) )
		return;

	/* Blit the erased area */
	if ( (SDL_cursor->sw_erased->flags & SDL_HWACCEL) == SDL_HWACCEL ) {
		blit_cursor = SDL_cursor->sw_erased->hw_blit;
	} else {
		blit_cursor = SDL_cursor->sw_erased->sw_blit;
	}
	blit_cursor(SDL_cursor->sw_erased, &updates[2], screen, &updates[0]);

	/* If we're not moving the cursor */
	if ( ! SDL_movingcursor ) {
		SDL_UpdateRects(screen, 1, &updates[0]);
	}
	SDL_cursorstate &= ~CURSOR_ONSCREEN;
}
