/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_endian.c,v 1.3 1998/08/18 07:03:50 slouken Exp $";
#endif

/* Functions for dynamically reading and writing endian-specific values */

#include <stdio.h>
#include <string.h>
#include "SDL_types.h"
#include "SDL_error.h"
#include "SDL_endian.h"

#define swap16(X)  ((X<<8)|(X>>8))
#define swap32(X)  ((X<<24)|((X<<8)&0x00FF0000)|((X>>8)&0x0000FF00)|(X>>24))

static int endian = 0;

int SDL_CalculateEndian (void)
{
	Uint16 value = 255;
	Uint8  valueb[2];

	if ( ! endian ) {
		memcpy(valueb, &value, 2);
		if ( (valueb[0] == 0) && (valueb[1] == 255) )
			endian = SDL_BIG_ENDIAN;
		else
		if ( (valueb[0] == 255) && (valueb[1] == 0) )
			endian = SDL_LIL_ENDIAN;
	}
	return(endian);
}

static Uint32 Read32(FILE *fp, int src, int dst)
{
	Uint32 value;

	if ( fread(&value, sizeof(Uint32), 1, fp) != 1 ) {
		SDL_SetError("EOF on datastream");
		return(0);
	}
	if ( src != dst )
		value = swap32(value);
	return(value);
}
static Uint32 Read16(FILE *fp, int src, int dst)
{
	Uint16 value;

	if ( fread(&value, sizeof(Uint16), 1, fp) != 1 ) {
		SDL_SetError("EOF on datastream");
		return(0);
	}
	if ( src != dst )
		value = swap16(value);
	return(value);
}

static int Write16(FILE *fp, Uint16 value, int src, int dst)
{
	if ( src != dst )
		value = swap16(value);
	if ( fwrite(&value, sizeof(Uint16), 1, fp) != 1 ) {
		SDL_SetError("Couldn't write value to datastream");
		return(0);
	}
	return(1);
}
static int Write32(FILE *fp, Uint32 value, int src, int dst)
{
	if ( src != dst )
		value = swap32(value);
	if ( fwrite(&value, sizeof(Uint32), 1, fp) != 1 ) {
		SDL_SetError("Couldn't write value to datastream");
		return(0);
	}
	return(1);
}

Uint16 SDL_SwapLE16 (Uint16 item)
{
	if ( endian != SDL_LIL_ENDIAN ) {
		return(swap16(item));
	} else {
		return(item);
	}
}
Uint16 SDL_SwapBE16 (Uint16 item)
{
	if ( endian != SDL_BIG_ENDIAN ) {
		return(swap16(item));
	} else {
		return(item);
	}
}
Uint32 SDL_SwapLE32 (Uint32 item)
{
	if ( endian != SDL_LIL_ENDIAN ) {
		return(swap32(item));
	} else {
		return(item);
	}
}
Uint32 SDL_SwapBE32 (Uint32 item)
{
	if ( endian != SDL_BIG_ENDIAN ) {
		return(swap32(item));
	} else {
		return(item);
	}
}

Uint16 SDL_ReadLE16 (FILE *fp)
{
	return(Read16(fp, SDL_LIL_ENDIAN, endian));
}
Uint16 SDL_ReadBE16 (FILE *fp)
{
	return(Read16(fp, SDL_BIG_ENDIAN, endian));
}
Uint32 SDL_ReadLE32 (FILE *fp)
{
	return(Read32(fp, SDL_LIL_ENDIAN, endian));
}
Uint32 SDL_ReadBE32 (FILE *fp)
{
	return(Read32(fp, SDL_BIG_ENDIAN, endian));
}

int SDL_WriteLE16 (FILE *fp, Uint16 value)
{
	return(Write16(fp, value, SDL_LIL_ENDIAN, endian));
}
int SDL_WriteBE16 (FILE *fp, Uint16 value)
{
	return(Write16(fp, value, SDL_BIG_ENDIAN, endian));
}
int SDL_WriteLE32 (FILE *fp, Uint32 value)
{
	return(Write32(fp, value, SDL_LIL_ENDIAN, endian));
}
int SDL_WriteBE32 (FILE *fp, Uint32 value)
{
	return(Write32(fp, value, SDL_BIG_ENDIAN, endian));
}

#ifdef TEST_ENDIAN
int main(int argc, char *argv[])
{
	SDL_CalculateEndian();
	if ( endian == SDL_LIL_ENDIAN )
		printf("This is a little endian machine\n"); 
	else
	if ( endian == SDL_BIG_ENDIAN )
		printf("This is a big endian machine\n"); 
	else
		printf("Unknown endian machine!\n");
	exit(0);
}
#endif
