/*
    SDL - Simple DirectMedia Layer
    Copyright (C) 1997, 1998  Sam Lantinga

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_events.c,v 1.5 1998/09/30 01:33:21 slouken Exp $";
#endif

/* General event handling code for SDL */

#include <stdio.h>
#include <string.h>

#include "SDL.h"
#include "SDL_mutex.h"
#include "SDL_events.h"
#include "SDL_events_c.h"
#include "SDL_sysevents.h"

#include "SDL_syswm.h"

/* Public data -- the event filter */
SDL_EventFilter SDL_EventOK = NULL;
Uint8 SDL_ProcessEvents[SDL_NUMEVENTS];

struct private_event
{
  SDL_Event pubpart;
  SDL_SysWMmsg privpart;
};

/* Private data -- event queue */
/* Note that we don't ever have more than one thread pushing events onto it,
   or one thread (possibly another thread) pulling events from it.
 */
#define MAXEVENTS	128
static struct {
	int	head;
	int	tail;
	int	quit;			/* Asyncronous Ctrl-C interrupt */
	struct private_event event[MAXEVENTS];
#ifdef THREADED_EVENTS
	SDL_mutex *wait;
#endif
} SDL_EventQ;

/* Public functions */

void SDL_StopEventLoop(void)
{
#ifdef THREADED_EVENTS
	/* Halt the event thread, if running */
	SDL_StopEventThread();
	SDL_DestroyMutex(SDL_EventQ.wait);
#endif
	/* Clean out EventQ */
	SDL_EventQ.head = 0;
	SDL_EventQ.tail = 0;
	SDL_EventQ.quit = 0;
}

/* This function (and associated calls) may be called more than once */
int SDL_StartEventLoop(void)
{
	int retcode;

	/* Initialize event handlers */
	retcode = 0;
	retcode += SDL_AppActiveInit();
	retcode += SDL_KeyboardInit();
	retcode += SDL_MouseInit();
	retcode += SDL_QuitInit();
	if ( retcode < 0 ) {
		/* We don't expect them to fail, but... */
		return(-1);
	}

	/* Clean out the event queue */
	SDL_EventQ.head = 0;
	SDL_EventQ.tail = 0;
	SDL_EventQ.quit = 0;

	/* No filter to start with, process most event types */
	SDL_EventOK = NULL;
	memset(SDL_ProcessEvents,SDL_ENABLE,sizeof(SDL_ProcessEvents));
	SDL_ProcessEvents[SDL_SYSWMEVENT] = SDL_IGNORE;

	/* Start up the event thread, and go! */
#ifdef THREADED_EVENTS
	SDL_EventQ.wait = SDL_CreateMutex();
	if ( SDL_EventQ.wait == NULL ) {
		return(-1);
	}
	SDL_mutexP(SDL_EventQ.wait);
	if ( SDL_StartEventThread() < 0 ) {
		SDL_StopEventLoop();
		return(-1);
	}
#endif
	return(0);
}

static void SDL_PushEventInternal(SDL_Event *event, SDL_SysWMmsg *message)
{
	int tail;

	tail = (SDL_EventQ.tail+1)%MAXEVENTS;
	if ( tail == SDL_EventQ.head ) {
		/* Overflow, drop event */
	} else {
		SDL_EventQ.event[SDL_EventQ.tail].pubpart = *event;
		if (message) {
			SDL_EventQ.event[SDL_EventQ.tail].privpart = *message;
		        SDL_EventQ.event[SDL_EventQ.tail].pubpart.syswm.msg =
			  &SDL_EventQ.event[SDL_EventQ.tail].privpart;
		}
		SDL_EventQ.tail = tail;
	}
#ifdef THREADED_EVENTS
	/* Release waiting app */
	SDL_mutexV(SDL_EventQ.wait);
#endif
	return;
}

void SDL_PushEvent(SDL_Event *event)
{
	SDL_PushEventInternal (event, 0);
}

static void SDL_PullEvent(SDL_Event *event)
{
	*event = SDL_EventQ.event[SDL_EventQ.head].pubpart;
	SDL_EventQ.head = (SDL_EventQ.head+1)%MAXEVENTS;
	return;
}

void SDL_QuitInterrupt(void)
{
	++SDL_EventQ.quit;
	SDL_SYS_QuitInterrupt();
}

int SDL_CheckQuit(SDL_Event *event)
{
	if ( SDL_EventQ.quit ) {
		if ( event ) {
			event->type = SDL_QUIT;
			--SDL_EventQ.quit;
		}
		return(1);
	}
	return(0);
}

/* Public functions */

#ifdef THREADED_EVENTS
void SDL_SYS_PollEvent(void)
{
	return;
}
int SDL_SYS_WaitEvent(void)
{
	SDL_mutexP(SDL_EventQ.wait);
	return(!SDL_EventQ.quit);
}
void SDL_SYS_QuitInterrupt(void)
{
	SDL_mutexV(SDL_EventQ.wait);
}
#endif /* THREADED_EVENTS */

int SDL_PollEvent (SDL_Event *event)
{
	if ( SDL_CheckQuit(event) ) {
		return(1);
	}
	if ( SDL_EventQ.head == SDL_EventQ.tail ) {
		SDL_SYS_PollEvent();
	}
	if ( SDL_EventQ.head == SDL_EventQ.tail ) {
		return(0);
	} else {
		if ( event ) {
#ifdef THREADED_EVENTS
			/* Decrement the event mutex */
			SDL_mutexP(SDL_EventQ.wait);
#endif
			SDL_PullEvent(event);
		}
		return(1);
	}
}

int SDL_WaitEvent (SDL_Event *event)
{
	/* Check for events already pending first */
	if ( SDL_PollEvent(event) ) {
		return(1);
	}

	/* If we get a quit interrupt, assume SDL_SYS_WaitEvent() returns 0 */
	while ( ! SDL_SYS_WaitEvent() ) {
		if ( SDL_CheckQuit(event) ) {
			return(1);
		}
		SDL_Delay(10);
	}
	if ( SDL_EventQ.head == SDL_EventQ.tail ) {
		/* SDL_SYS_WaitEvent() returned non-zero, but no event! */
		return(0);
	} else {
		if ( event ) {
#ifdef THREADED_EVENTS
			/* The event mutex decremented by SDL_SYS_WaitEvent() */
#endif
			SDL_PullEvent(event);
		}
		return(1);
	}
}

void SDL_SetEventFilter (SDL_EventFilter filter)
{
	SDL_Event bitbucket;

	/* Set filter and discard pending events */
	SDL_EventOK = filter;
	while ( SDL_PollEvent(&bitbucket) )
		;
}

Uint8 SDL_EventState (Uint8 type, int state)
{
	SDL_Event bitbucket;
	Uint8 current_state;

	current_state = SDL_ProcessEvents[type];
	switch (state) {
		case SDL_IGNORE:
		case SDL_ENABLE:
			/* Set state and discard pending events */
			SDL_ProcessEvents[type] = state;
			while ( SDL_PollEvent(&bitbucket) )
				;
			break;
		default:
			/* Querying state? */
			break;
	}
	return(current_state);
}

/* This is a generic event handler.
 */
int SDL_PrivateSysWMEvent(SDL_SysWMmsg *message)
{
	int posted;

	posted = 0;
	if ( SDL_ProcessEvents[SDL_SYSWMEVENT] == SDL_ENABLE ) {
		SDL_Event event;
		event.type = SDL_SYSWMEVENT;
		event.syswm.msg = message;
		if ( (SDL_EventOK == NULL) || (*SDL_EventOK)(&event) ) {
			posted = 1;
			SDL_PushEventInternal(&event, message);
		}
	}
	/* Update internal event state */
	return(posted);
}
