/*
    MIXERLIB:  An audio mixer library based on the SDL library
    Copyright (C) 1997  Sam Lantinga

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Sam Lantinga
    5635-34 Springhouse Dr.
    Pleasanton, CA 94588 (USA)
    slouken@devolution.com
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_mutex.h"
#include "SDL_endian.h"

#include "mixer.h"
#include "mikmod/mikmod.h"

static int audio_opened = 0;

static SDL_AudioSpec mixer;
static SDL_mutex *mixer_lock;
static struct _Mix_Channel {
	int playing;
	Uint8 *samples;
} channel[MIX_CHANNELS];

static int music_playing = 0;
static int music_swap8;
static int music_swap16;
struct _Mix_Music {
	UNIMOD *module;
};
#define MUSIC_VOL	64		/* Music volume 0-64 */


/* Mixing function */
static void mix_channels(void *udata, Uint8 *stream, int len)
{
	struct _Mix_Channel mixable[MIX_CHANNELS];
	int i;

	/* Grab the channels we need to mix */
	SDL_mutexP(mixer_lock);
	for ( i=0; i<MIX_CHANNELS; ++i ) {
		mixable[i].samples = channel[i].samples;
		mixable[i].playing = channel[i].playing;
		if ( mixable[i].playing > len ) {
			mixable[i].playing = len;
		}
		channel[i].samples += mixable[i].playing;
		channel[i].playing -= mixable[i].playing;
	}
	SDL_mutexV(mixer_lock);

	/* Mix them all */
	if ( music_playing ) {
		VC_WriteBytes((SBYTE *)stream, len);
		if ( music_swap8 ) {
			Uint8 *dst;

			dst = stream;
			for ( i=len; i; --i ) {
				*dst++ ^= 0x80;
			}
		} else
		if ( music_swap16 ) {
			Uint8 *dst, tmp;

			dst = stream;
			for ( i=(len/2); i; --i ) {
				tmp = dst[0];
				dst[0] = dst[1];
				dst[1] = tmp;
				dst += 2;
			}
		}
	}
	for ( i=0; i<MIX_CHANNELS; ++i ) {
		SDL_MixAudio(stream, mixable[i].samples, mixable[i].playing);
	}
}

/* Open the mixer with a certain desired audio format */
int Mix_OpenAudio(Uint32 frequency, Uint16 format, Uint16 chunksize)
{
	int i;
	SDL_AudioSpec desired;

	/* Set the desired format and frequency */
	desired.freq = frequency;
	desired.format = format;
	desired.samples = chunksize;
	desired.callback = mix_channels;
	desired.userdata = NULL;

	/* Accept nearly any audio format */
	if ( SDL_OpenAudio(&desired, &mixer) < 0 ) {
		return(-1);
	}

	/* Create the channel lock mutex */
	mixer_lock = SDL_CreateMutex();
	if ( mixer_lock == NULL ) {
		SDL_CloseAudio();
		SDL_SetError("Unable to create mixer lock");
		return(-1);
	}

	/* Set the MikMod music format */
	music_swap8 = 0;
	music_swap16 = 0;
	switch (mixer.format&~AUDIO_STEREO) {

		case AUDIO_U8:
		case AUDIO_S8: {
			if ( (mixer.format&~AUDIO_STEREO) == AUDIO_S8 ) {
				music_swap8 = 1;
			}
			md_mode = 0;
		}
		break;

		case AUDIO_S16LSB:
		case AUDIO_S16MSB: {
			/* See if we need to correct MikMod mixing */
			if ( (((mixer.format&~AUDIO_STEREO) == AUDIO_S16LSB) &&
			      (SDL_CalculateEndian() != SDL_LIL_ENDIAN)) ||
			     (((mixer.format&~AUDIO_STEREO) == AUDIO_S16MSB) &&
			      (SDL_CalculateEndian() != SDL_BIG_ENDIAN)) ) {
				music_swap16 = 1;
			}
			md_mode = DMODE_16BITS;
		}
		break;

		default: {
			SDL_DestroyMutex(mixer_lock);
			SDL_CloseAudio();
			SDL_SetError("Unknown hardware audio format");
			return(-1);
		}
	}
	if ( (mixer.format & AUDIO_STEREO) == AUDIO_STEREO ) {
		md_mode |= DMODE_STEREO;
	}
	md_mixfreq     = mixer.freq;
	md_device      = 0;
	md_volume      = 96;
	md_musicvolume = 128;
	md_sndfxvolume = 128;
	md_pansep      = 128;
	md_reverb      = 0;
	MikMod_RegisterAllLoaders();
	MikMod_RegisterAllDrivers();
	if ( MikMod_Init() ) {
		SDL_DestroyMutex(mixer_lock);
		SDL_CloseAudio();
		SDL_SetError("%s", _mm_errmsg[_mm_errno]);
		return(-1);
	}
	music_playing = 0;
	
	/* Clear out the audio channels */
	for ( i=0; i<MIX_CHANNELS; ++i ) {
		channel[i].playing = 0;
	}
	audio_opened = 1;
	SDL_PauseAudio(0);
	return(0);
}

/* Return the actual mixer parameters */
void Mix_QuerySpec(Uint32 *frequency, Uint16 *format)
{
	if ( audio_opened ) {
		*frequency = mixer.freq;
		*format = mixer.format;
	}
}

static void PrintFormat(char *title, SDL_AudioSpec *fmt)
{
	printf("%s: %d bit %s audio (%s) at %u Hz\n", title, (fmt->format&0xFF),
			(fmt->format&0x8000) ? "signed" : "unsigned",
			(fmt->format&AUDIO_STEREO) ? "stereo" : "mono",
								fmt->freq);
}

/* Load a wave file */
Mix_Chunk *Mix_LoadWAV(const char *file)
{
	Mix_Chunk *chunk;
	SDL_AudioSpec wavespec;
	SDL_AudioCVT wavecvt;
	Uint8 *cvtbuf;

	/* Make sure audio has been opened */
	if ( ! audio_opened ) {
		SDL_SetError("Audio device hasn't been opened");
		return(NULL);
	}

	/* Allocate the chunk memory */
	chunk = (Mix_Chunk *)malloc(sizeof(Mix_Chunk));
	if ( chunk == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}

	/* Load the WAV file into the chunk */
	if ( SDL_LoadWAV(file, &wavespec, (Uint8 **)&chunk->abuf, &chunk->alen)
								 == NULL ) {
		free(chunk);
		return(NULL);
	}
#if 0
	PrintFormat("Audio device", &mixer);
	PrintFormat("-- Wave file", &wavespec);
#endif

	/* Build the audio converter and create conversion buffers */
	if ( SDL_BuildAudioCVT(&wavecvt, wavespec.format, mixer.format,
					wavespec.freq, mixer.freq) < 0 ) {
		free(chunk->abuf);
		free(chunk);
		return(NULL);
	}
	if ( wavecvt.len_mult == 1 ) {
		wavecvt.len = chunk->alen;
		wavecvt.buf = chunk->abuf;
	} else {
		wavecvt.len = chunk->alen;
		wavecvt.buf = (Uint8 *)malloc(wavecvt.len*wavecvt.len_mult);
		if ( wavecvt.buf == NULL ) {
			SDL_SetError("Out of memory");
			free(chunk->abuf);
			free(chunk);
			return(NULL);
		}
		memcpy(wavecvt.buf, chunk->abuf, chunk->alen);
		free(chunk->abuf);
	}

	/* Run the audio converter */
	if ( SDL_ConvertAudio(&wavecvt) < 0 ) {
		free(wavecvt.buf);
		free(chunk);
		return(NULL);
	}
	chunk->abuf = wavecvt.buf;
	chunk->alen = wavecvt.len_cvt;
	return(chunk);
}
Mix_Music *Mix_LoadMUS(const char *file)
{
	Mix_Music *music;

	music = (Mix_Music *)malloc(sizeof(Mix_Music));
	if ( music == NULL ) {
		SDL_SetError("Out of memory");
		return(NULL);
	}
	music->module = MikMod_LoadSong((char *)file, 64);
	if ( music->module == NULL ) {
		SDL_SetError("%s", _mm_errmsg[_mm_errno]);
		free(music);
		return(NULL);
	}
	return(music);
}

/* Free an audio chunk previously loaded */
void Mix_FreeChunk(Mix_Chunk *chunk)
{
	/* Caution -- if the chunk is playing, the mixer will crash */
	if ( chunk ) {
		free(chunk->abuf);
		free(chunk);
	}
}
void Mix_FreeMusic(Mix_Music *music)
{
	/* Caution -- if the music is playing, the mixer will crash */
	if ( music ) {
		MikMod_FreeSong(music->module);
		free(music);
	}
}

/* Play an audio chunk on a specific channel.
   If the specified channel is -1, play on the first free channel.
   Returns which channel was used to play the sound.
*/
int Mix_PlayChannel(int which, Mix_Chunk *chunk)
{
	int i;

	/* Don't play null pointers :-) */
	if ( chunk == NULL ) {
		return(-1);
	}

	/* If which is -1, play on the first free channel */
	if ( which == -1 ) {
		for ( i=0; i<MIX_CHANNELS; ++i ) {
			if ( channel[i].playing <= 0 )
				break;
		}
		if ( i == MIX_CHANNELS ) {
			return(-1);
		}
		which = i;
	}

	/* Queue up the audio data for this channel */
	SDL_mutexP(mixer_lock);
	channel[which].samples = chunk->abuf;
	channel[which].playing = chunk->alen;
	SDL_mutexV(mixer_lock);

	/* Return the channel on which the sound is being played */
	return(which);
}
int Mix_PlayMusic(Mix_Music *music)
{
	/* Don't play null pointers :-) */
	if ( music == NULL ) {
		return(-1);
	}
	Player_Start(music->module);
	music_playing = 1;
	return(0);
}

/* Halt playing of a particular channel */
int Mix_HaltChannel(int which)
{
	int i;

	if ( which == -1 ) {
		for ( i=0; i<MIX_CHANNELS; ++i ) {
			Mix_HaltChannel(i);
		}
	} else {
		SDL_mutexP(mixer_lock);
		if ( channel[which].playing > 0 ) {
			channel[which].playing = 0;
		}
		SDL_mutexV(mixer_lock);
	}
	return(0);
}
int Mix_HaltMusic(void)
{
	if ( music_playing ) {
		Player_Stop();
		music_playing = 0;
	}
	return(0);
}

/* Check the status of a specific channel.
   If the specified channel is -1, check all channels.
*/
int Mix_Playing(int which)
{
	int status;

	status = 0;
	if ( which == -1 ) {
		int i;

		for ( i=0; i<MIX_CHANNELS; ++i ) {
			if ( channel[i].playing > 0 ) {
				++status;
			}
		}
	} else {
		if ( channel[which].playing > 0 ) {
			++status;
		}
	}
	return(status);
}
int Mix_PlayingMusic(void)
{
	if ( music_playing ) {
		return(Player_Active());
	}
	return(0);
}

/* Close the mixer, halting all playing audio */
void Mix_CloseAudio(void)
{
	int i;

	if ( audio_opened ) {
		Mix_HaltMusic();
		MikMod_Exit();
		for ( i=0; i<MIX_CHANNELS; ++i ) {
			Mix_HaltChannel(i);
		}
		SDL_CloseAudio();
		SDL_DestroyMutex(mixer_lock);
		audio_opened = 0;
	}
}
